/*
** Module   : AbstractButton.java
** Abstract : common button like widgets implementation
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20140610 Initial implementation.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of a "abstract button" with a text label. This is the place to handle
 * features common for all button like objects, e.g. buttons, toggle-boxes.  This is one line
 * height button like widget.
 */
public abstract class AbstractButton
extends AbstractWidget
implements KeyListener,
           FocusListener
{
   /** Button label. */
   private String textLabel;

   /** Widget in focus flag. */
   protected boolean widgetFocused = false;
   
   /** Previous cursor visibility flag to restore on focus lost. */
   private boolean curvis = false;
   
   /**
    * Default constructor.
    */
   public AbstractButton()
   {
      super();
   }

   // AbstractWidget overrides start
   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Draw the button implementation. Called by this AbstarctButton's parent container.
    */
   @Override
   public abstract void draw();

   /**
    * Check if widget needs in cursor to be on when having the input focus.  For toggle-box widget
    * this method should always return <code>false</code> because toggle-box never needs cursor
    * visibility, even when it is required to force the cursor to be on.
    * 
    * @return  <code>false</code> unconditionally.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return false;
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }
   // AbstractWidget overrides end
   
   // Common button-like widgets public and private code
   /**
    * Processes key events occurring on this object by dispatching them
    * to any registered KeyListener objects.
    *
    * @param    keyEvent
    *           The event to be handled.
    */
   protected void processKeyEventCommon(KeyInput keyEvent)
   {
      // use arrow keys to move focus between widgets (only left/right work
      // in the alert box case)
      if (parent() instanceof AlertBox)
      {
         setActionCode(keyEvent);
      }
      else
      {
         if (handleCursorKeys(keyEvent.actionCode(), (Frame) UiUtils.locateFrame(this)))
            return;
      }

      handleBasicEvents(keyEvent);
   }

   /**
    * Process basic events - TAB, BACK-TAB and ENTER key event.
    * 
    * @param   event
    *          Key event to process.
    */
   protected abstract void handleBasicEvents(KeyInput event);
   
   /**
    * Handle cursor movement keys.
    * 
    * @param   key
    *          Key code.
    * @param   frame
    *          Parent frame.
    * 
    * @return  <code>true</code> if any cursor key was found.
    */
   protected boolean handleCursorKeys(int key, Frame frame)
   {
      boolean returnNow = false;
      
      if (key == Keyboard.KA_TAB)
      {
         returnNow = nextFocus();
      }
      else if (key == Keyboard.KA_BACK_TAB)
      {
         returnNow = prevFocus();
      }
      else if (key == Keyboard.KA_CURSOR_LEFT)
      {
         // the left most widget ignores this condition
         if (!frame.isLeftMostWidget(this))
         {
            returnNow = prevFocus();
         }
      }
      else if (key == Keyboard.KA_CURSOR_UP)
      {
         returnNow = moveAboveWidget(frame);
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         // the right most widget ignores this condition
         if (!frame.isRightMostWidget(this))
         {
            returnNow = nextFocus();
         }
      }
      else if (key == Keyboard.KA_CURSOR_DOWN)
      {
         returnNow = moveBelowWidget(frame);
      }
      return returnNow;
   }

   /**
    * Move focus to widget below in specified frame.
    * 
    * @param   frame
    *          Parent frame.
    *          
    * @return  <code>true</code>
    */
   private boolean moveBelowWidget(Frame frame)
   {
      frame.moveBelowWidget(this, 0);
      refresh();
      return true;
   }

   /**
    * Move focus to widget above in specified frame.
    * 
    * @param   frame
    *          Parent frame.
    *          
    * @return  <code>true</code>
    */
   private boolean moveAboveWidget(Frame frame)
   {
      frame.moveAboveWidget(this, 0);
      refresh();
      return true;
   }

   /**
    * Move focus to previous widget in Tab order.
    * 
    * @return  <code>true</code>
    */
   private boolean prevFocus()
   {
      ThinClient.previousFocus(this);
      refresh();
      return true;
   }

   /**
    * Move focus to next widget in Tab order.
    * 
    * @return  <code>true</code>
    */
   private boolean nextFocus()
   {
      ThinClient.nextFocus(this);
      refresh();
      return true;
   }

   /**
    * Refresh widget at the screen.
    */
   protected void refresh()
   {
      draw();
      requestSync();
   }

   /**
    * Prepare event for processing.
    *   
    * @param   ki
    *          Input event.
    */
   protected void setActionCode(KeyInput ki)
   {
      int key = ki.actionCode();
      // alert box processing has no need to generate entry/leave events 
      // so we take a very simple approach
      if (key == Keyboard.KA_CURSOR_LEFT)
      {
         ki.setActionCode(Keyboard.KA_BACK_TAB);
         ki.setKeyCode(Key.VK_BACK_TAB);
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         ki.setActionCode(Keyboard.KA_TAB);
         ki.setKeyCode(Key.VK_TAB);         
      }
   }
   
   /**
    * Get label text.
    * 
    * @return  label text.
    */
   public String textLabel()
   {
      return textLabel;
   }

   /**
    * Set the new text for the component.
    * 
    * @param   text
    *          New text value.
    */
   public abstract void setText(String text);

   /**
    * Set button label.
    * 
    * @param   label
    *          Label text.
    */
   public void setTextLabel(String label)
   {
      this.textLabel = label;
   }

   // KeyListener interface implementation start
   /**
    * This method is invoked when key is pressed.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onKeyPressed(KeyInput event)
   {
   }

   /**
    * This method is invoked when key is typed.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onKeyTyped(KeyInput event)
   {
   }
   // KeyListener interface implementation end
   
   // FocusListener interface implementation start
   /**
    * This is called when the toggle-box takes the input focus. When the help string is defined
    * for the toggle-box it should be displayed at the window status line.
    *
    * @param   event
    *          The focus event to be processed.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      // no need in code below if widget already focused
      if (widgetFocused)
      {
         return;
      }
      
      // force the cursor off since it should not display while we have focus
      curvis = screen().setCursorStatus(false);

      // mark widget as gained input focus      
      widgetFocused = true;
      
      // even for consecutive FOCUS_GAINED events, the HELP text must be set
      // for this toggle-box, because the HELP message-text may have been changed
      // for another widget (i.e. on a nested UPDATE)
      
      // draw the help string for focused toggle-box
      ThinClient.setHelp(config().getHelp());
   }
   
   /**
    * This is called when the button looses the input focus.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusLost(FocusEvent focusEvent)
   {
      // no need in code below if widget already lost the focus
      if (!widgetFocused)
      {
         return;
      }
      
      widgetFocused = false;
      
      // restore the cursor visibility that existed when we gained focus 
      screen().setCursorStatus(curvis);
   }
   // FocusListener interface implementation end
}
